/*
*
* Solver: QpExtSolCBC.cpp -- Copyright (c) 2010-2017 Jan Wolf
*
* Permission is hereby granted, free of charge, to any person obtaining a
* copy of this software and associated documentation files (the
* "Software"), to deal in the Software without restriction, including
* without limitation the rights to use, copy, modify, merge, publish,
* distribute, sublicense, and/or sell copies of the Software, and to
* permit persons to whom the Software is furnished to do so, subject to
* the following conditions:
*
* The above copyright notice and this permission notice shall be included
* in all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
* OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
* NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
* LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
* OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
* WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include "ExternSolvers/QpExtSolCBC.hpp"

#ifdef COMPILE_WITH_CBC

namespace extSol {

const std::string QpExtSolCBC::LOG_TAG = "QpExternSolverCBC";

QpExtSolCBC::QpExtSolCBC() :
		QpExternSolver(CBC, DEFAULT),solver(),model(solver),cm(),origConstraints(0){
	this->adaptToSolverMode(this->sMode);
}

QpExtSolCBC::QpExtSolCBC(const data::Qlp& qlp) :
		QpExternSolver(CBC, DEFAULT),solver(),model(solver),cm(),origConstraints(0){
	this->init(qlp);
	this->adaptToSolverMode(sMode);
}

QpExtSolCBC::QpExtSolCBC(const data::QpObjFunc& obj,
		const std::vector<data::QpVar>& vars, const data::QpSparseMatrix& mat,
		const std::vector<data::QpRhs>& rhs) :
		QpExternSolver(CBC, DEFAULT),solver(),model(solver),cm(),origConstraints(0){
	this->init(obj, vars, mat, rhs);
	this->adaptToSolverMode(this->sMode);
}

QpExtSolCBC::QpExtSolCBC(const std::string&lpfile) :
		QpExternSolver(CBC, DEFAULT),solver(),model(solver),cm(),origConstraints(0){
	this->init(lpfile);
	this->adaptToSolverMode(this->sMode);
}

QpExtSolCBC::~QpExtSolCBC() {
}

void QpExtSolCBC::clear(){
	solver = OsiClpSolverInterface();
	model = CbcModel(solver);
#ifdef DISABLE_CBC_OUTPUT
	cm.setLogLevel(0);
	model.passInMessageHandler(&cm);
#endif
}

void * QpExtSolCBC::getSolverEnv() {
	return NULL;
}

void * QpExtSolCBC::getSolverModel() {
	return NULL;
}

void QpExtSolCBC::init(const data::Qlp& qlp) {
	std::vector<std::vector<data::IndexedElement> > matrix;
	qlp.getCoeffMatrix(matrix);
	this->init(qlp.getObjectiveFunction(), qlp.getQuantifiedVariables(), matrix,
			qlp.getRhsVec());
}

void QpExtSolCBC::init(const data::Qlp& qlp, data::QpRhs::Responsibility resp) {
	std::vector<std::vector<data::IndexedElement> > matrix;
	qlp.getCoeffMatrixByResp(matrix,resp);
	this->init(qlp.getObjectiveFunction(), qlp.getQuantifiedVariables(), matrix,
	qlp.getRhsVecByResp(resp));
}

void QpExtSolCBC::init(const data::QpObjFunc& obj,
		const std::vector<data::QpVar>& vars, const data::QpSparseMatrix& mat,
		const std::vector<data::QpRhs>& rhs) {
	this->clear();
	std::vector<int> indices(0);
	std::vector<double> values(0);

	model.solver()->setObjSense(
			obj.getObjective() == data::QpObjFunc::min ? 1 : -1);

	if(!obj.getOffset().isZero())
		throw utils::ExternSolverException("QpExtSolCBC::init( ... ) --> CBC cannot handle objective function offset");

	utils::Logger::globalLog(utils::LOG_INFO, "Test::debugExternSolver()", "Adding Cols");
	for (unsigned int i = 0; i < vars.size(); i++) {
		model.solver()->addCol(0, indices.data(), values.data(),
				vars[i].getLowerBound().asDouble(),
				vars[i].getUpperBound().asDouble(), obj[i].asDouble());
		if (vars[i].getNumberSystem() != data::QpVar::real){
			model.solver()->setInteger(i);
		}
	}


	utils::Logger::globalLog(utils::LOG_INFO, "Test::debugExternSolver()", "Adding Rows");
	int size;
	double lb, ub;
	for (unsigned int i = 0; i < mat.size(); i++) {
		size = mat[i].size();
		lb = -model.solver()->getInfinity();
		ub = model.solver()->getInfinity();
		if (rhs[i].getRatioSign() == data::QpRhs::equal) {
			lb = ub = rhs[i].getValue().asDouble();
		} else if (rhs[i].getRatioSign() == data::QpRhs::smallerThanOrEqual) {
			ub = rhs[i].getValue().asDouble();
		} else {
			lb = rhs[i].getValue().asDouble();
		}
		indices.resize(size);
		values.resize(size);
		for (unsigned int j = 0; j < size; j++) {
			indices[j] = mat[i][j].index;
			values[j] = mat[i][j].value.asDouble();
		}
		model.solver()->addRow(size, indices.data(), values.data(), lb, ub);
	}
	utils::Logger::globalLog(utils::LOG_INFO, "Test::debugExternSolver()", "Getting RowCount");
	this->origConstraints = this->getRowCount();
	utils::Logger::globalLog(utils::LOG_INFO, "Test::debugExternSolver()", "Done");
}

void QpExtSolCBC::init(const std::string& lpfile) {
	this->readFromFile(lpfile);
}

unsigned int QpExtSolCBC::getVariableCount() const {
	return model.solver()->getNumCols();
}

unsigned int QpExtSolCBC::getRowCount() const {
	return model.solver()->getNumRows();
}

unsigned int QpExtSolCBC::getCutCount() const {
	return this->origConstraints - model.solver()->getNumRows();
}

unsigned int QpExtSolCBC::getNonZeros() const {
	return model.solver()->getNumElements();
}

void QpExtSolCBC::readFromFile(const std::string& lpfile) {
	this->clear();
	model.solver()->readLp(lpfile.c_str());
	this->origConstraints = this->getRowCount();
}

void QpExtSolCBC::writeToFile(const std::string& path,
		const std::string& name) {
	model.solver()->writeLp((path + name).c_str());
//	//TODO je nach Endung mps oder lp schreiben
}

void QpExtSolCBC::getBase(extSol::QpExternSolver::QpExtSolBase& base) const {
//	base.variables.resize(this->getVariableCount(),
//			extSol::QpExternSolver::NotABasicStatus);
//	base.constraints.resize(this->getRowCount(),
//			extSol::QpExternSolver::NotABasicStatus);
//	solver->getBasisStatus(base.variables.data(),
//			base.constraints.data());
//
//	//	   This method returns status as integer codes:
//	//	      <li> 0: free
//	//	      <li> 1: basic
//	//	      <li> 2: nonbasic at upper bound
//	//	      <li> 3: nonbasic at lower bound
//	for (unsigned int i = 0; i < base.variables.size(); i++) {
//		switch (base.variables[i]) {
//		case 0:
//			base.variables[i] = extSol::QpExternSolver::NotABasicStatus;
//			break;
//		case 1:
//			base.variables[i] = extSol::QpExternSolver::Basic;
//			break;
//		case 2:
//			base.variables[i] = extSol::QpExternSolver::AtUpper;
//			break;
//		case 3:
//			base.variables[i] = extSol::QpExternSolver::AtLower;
//			break;
//		default:
//			throw utils::ExternSolverException(
//					" QpExtSolCBC::getBase(...) --> Unsupported Status in base.variables: "
//							+ utils::ToolBox::convertToString(
//									base.variables[i]));
//		}
//	}
//
//	for (unsigned int i = 0; i < base.constraints.size(); i++) {
//		switch (base.constraints[i]) {
//		case 0:
//			base.constraints[i] = extSol::QpExternSolver::NotABasicStatus;
//			break;
//		case 1:
//			base.constraints[i] = extSol::QpExternSolver::Basic;
//			break;
//		case 2:
//			base.constraints[i] = extSol::QpExternSolver::AtUpper;
//			break;
//		case 3:
//			base.constraints[i] = extSol::QpExternSolver::AtLower;
//			break;
//		default:
//			throw utils::ExternSolverException(
//					" QpExtSolCBC::getBase(...) --> Unsupported Status in base.constraints: "
//							+ utils::ToolBox::convertToString(
//									base.constraints[i]));
//		}
//	}
}

void QpExtSolCBC::setBase(extSol::QpExternSolver::QpExtSolBase& base) {
//
		if ((base.constraints.size() != this->getRowCount()) || (base.variables.size() != this->getVariableCount())) {
	throw utils::ExternSolverException("QpExtSolCplexC::setBase(const extSol::QpExternSolver::QpExtSolBase& base) -> (base.constraints.size() != this->getRowCount()) || (base.variables.size()!=this->getVariableCount())");
}

//
//	CoinWarmStartBasis* ws = new CoinWarmStartBasis;
//	ws->setSize( base.variables.size(), base.constraints.size() );
//
//	for (unsigned int i = 0; i < base.variables.size(); i++) {
//		switch (base.variables[i]) {
//		case -1:
//			 ws->setStructStatus( i, CoinWarmStartBasis::isFree );
//			break;
//		case 0:
//			 ws->setStructStatus( i, CoinWarmStartBasis::atLowerBound );
//			break;
//		case 1:
//			 ws->setStructStatus( i, CoinWarmStartBasis::basic );
//			break;
//		case 2:
//			 ws->setStructStatus( i, CoinWarmStartBasis::atUpperBound );
//			break;
//		case 3:
//			 ws->setStructStatus( i, CoinWarmStartBasis::isFree );
//			break;
//		default:
//			throw utils::ExternSolverException(
//					" QpExtSolCBC::getSase(...) --> Unsupported Status in base.variables: "
//							+ utils::ToolBox::convertToString(
//									base.variables[i]));
//		}
//	}
//
//	for (unsigned int i = 0; i < base.constraints.size(); i++) {
//			switch (base.constraints[i]) {
//			case -1:
//				ws->setArtifStatus( i, CoinWarmStartBasis::isFree );
//				break;
//			case 0:
//				ws->setArtifStatus( i, CoinWarmStartBasis::atLowerBound );
//				break;
//			case 1:
//				ws->setArtifStatus( i, CoinWarmStartBasis::basic );
//				break;
//			case 2:
//				ws->setArtifStatus( i, CoinWarmStartBasis::atUpperBound );
//				break;
//			case 3:
//				ws->setArtifStatus( i, CoinWarmStartBasis::isFree );
//				break;
//			default:
//				throw utils::ExternSolverException(
//						" QpExtSolCBC::setBase(...) --> Unsupported Status in base.constraints: "
//								+ utils::ToolBox::convertToString(
//										base.constraints[i]));
//			}
//		}
//
//	solver->setWarmStart(ws);
//	delete ws;
//	return;
//
//
//	//	   This method returns status as integer codes:
//	//	      <li> 0: free
//	//	      <li> 1: basic
//	//	      <li> 2: nonbasic at upper bound
//	//	      <li> 3: nonbasic at lower bound
//	//
//	//		typedef enum {
//	//			NotABasicStatus = -1, AtLower = 0, Basic = 1, AtUpper = 2, FreeOrSuperbasic = 3
//	//		} QpExtSolBasisStatus;
//	//		typedef std::vector<int> QpExtSolBasisStatusArray;
//	//
//
//	std::vector<int> variables(base.variables.size());
//	std::vector<int> constraints(base.constraints.size());
//
//	for (unsigned int i = 0; i < base.variables.size(); i++) {
//		switch (base.variables[i]) {
//		case -1:
//			variables[i] = 0;
//			break;
//		case 0:
//			variables[i] = 3;
//			break;
//		case 1:
//			variables[i] = 1;
//			break;
//		case 2:
//			variables[i] = 2;
//			break;
//		case 3:
//			variables[i] = 0;
//			break;
//		default:
//			throw utils::ExternSolverException(
//					" QpExtSolCBC::getSase(...) --> Unsupported Status in base.variables: "
//							+ utils::ToolBox::convertToString(
//									base.variables[i]));
//		}
//	}
//
//	for (unsigned int i = 0; i < base.constraints.size(); i++) {
//		switch (base.constraints[i]) {
//		case -1:
//			constraints[i] = 0;
//			break;
//		case 0:
//			constraints[i] = 3;
//			break;
//		case 1:
//			constraints[i] = 1;
//			break;
//		case 2:
//			constraints[i] = 2;
//			break;
//		case 3:
//			constraints[i] = 0;
//			break;
//		default:
//			throw utils::ExternSolverException(
//					" QpExtSolCBC::setBase(...) --> Unsupported Status in base.constraints: "
//							+ utils::ToolBox::convertToString(
//									base.constraints[i]));
//		}
//	}
//	// Not implemented
//	// solver->setBasisStatus(base.variables.data(),
//	// base.constraints.data());
}

void QpExtSolCBC::setRayGuess(const std::vector<data::QpNum>& rayGuess) {/*not implemented, only needed for ToSimplex*/

}

void QpExtSolCBC::adaptToSolverMode(QpExtSolSolverMode m) {

}

void QpExtSolCBC::setParameters(const ExtSolverParameters& p) {
}


void QpExtSolCBC::setVarLB(unsigned int i, const data::QpNum& lb) {
	this->model.solver()->setColLower(i, lb.asDouble());
}

void QpExtSolCBC::setVarUB(unsigned int i, const data::QpNum& ub) {
	this->model.solver()->setColUpper(i, ub.asDouble());
}

void QpExtSolCBC::changeRhsElement(unsigned int i, const data::QpNum& v) {
	if (this->model.solver()->getRowLower()[i] != -this->model.solver()->getInfinity())
		this->model.solver()->setRowLower(i, v.asDouble());
	if (this->model.solver()->getRowUpper()[i] != this->model.solver()->getInfinity())
		this->model.solver()->setRowUpper(i, v.asDouble());
}

void QpExtSolCBC::changeRhsElements(const std::vector<unsigned int>& indices, const std::vector<data::QpNum>& values) {
	for (unsigned int i = 0; i < indices.size(); i++)
		this->changeRhsElement(indices[i], values[i]);
}

void QpExtSolCBC::changeRhsElements(const std::vector<int>& indices, const std::vector<double>& values) {
    for (unsigned int i = 0; i < indices.size(); i++)
	    this->changeRhsElement((unsigned int)indices[i], data::QpNum(values[i]));
}

extSol::QpExternSolver::QpExtSolSolutionStatus QpExtSolCBC::getSolutionStatus() {
	if (model.isProvenOptimal())
		return extSol::QpExternSolver::OPTIMAL;
	if (model.isProvenInfeasible())
		return extSol::QpExternSolver::INFEASIBLE;
	if (model.isProvenDualInfeasible())
		return extSol::QpExternSolver::INForUNB;
	return extSol::QpExternSolver::ERROR;
}

extSol::QpExternSolver::QpExtSolSolutionStatus QpExtSolCBC::solve(
		unsigned int itLimit, unsigned int timeLimit) {

	this->model.setMaximumSeconds(timeLimit);
	CbcMain0( this->model);
	const char * argv[] = { "cbc", "-log", "3", "-solve", "-quit" };
	CbcMain1( 5, argv, this->model );
	//this->model.branchAndBound();
	return this->getSolutionStatus();
}

data::QpNum QpExtSolCBC::getObjValue() {
	return this->model.getSolverObjValue();
}

void QpExtSolCBC::getValues(std::vector<data::QpNum>& values) {
	const double* v = this->model.getCbcColSolution();
	unsigned int vars = this->getVariableCount();
	values.resize(vars);
	for (unsigned int i = 0; i < vars; i++)
		values[i] = v[i];
}

void QpExtSolCBC::getDuals(std::vector<data::QpNum>& duals) {
//	const double* v = this->solver->getRowPrice();
//	unsigned int cons = this->getRowCount();
//	duals.resize(cons);
//	for (unsigned int i = 0; i < cons; i++)
//		duals[i] = v[i];
}

void QpExtSolCBC::getReducedCosts(std::vector<data::QpNum>& reduced) {
//	const double* v = this->solver->getReducedCost();
//	unsigned int vars = this->getVariableCount();
//	reduced.resize(vars);
//	for (unsigned int i = 0; i < vars; i++)
//		reduced[i] = v[i];
}

void QpExtSolCBC::getDualFarkas(std::vector<data::QpNum>& farkas) {
//	unsigned int cons = this->getRowCount();
//	if (!cons) {
//		farkas.clear();
//		return;
//	}
//	std::vector<double *> v = solver->getDualRays(1, false);
//	farkas.resize(cons);
//	for (unsigned int i = 0; i < cons; i++) {
//		farkas[i] = v[0][i];
//	}
//
//	for (unsigned int i = 0; i < v.size(); i++)
//		delete v[i];
}

void QpExtSolCBC::getExtendedDuals(std::vector<data::QpNum>& extDuals) {
//	unsigned int vars = this->getVariableCount();
//	unsigned int cons = this->getRowCount();
//	this->getDuals(extDuals);
//	extDuals.resize(cons + 2 * vars, 0);
//	std::vector<data::QpNum> tmpSolParts;
//	this->getReducedCosts(tmpSolParts);
//	extSol::QpExternSolver::QpExtSolBase base;
//	this->getBase(base);
//	for (unsigned int i = 0, index = 0; i < tmpSolParts.size(); i++) {
//		if (!tmpSolParts[i].isZero()) {
//			if (base.variables[i] == extSol::QpExternSolver::AtLower) {
//				index = i;
//			} else if (base.variables[i] == extSol::QpExternSolver::AtUpper) {
//				index = vars + i;
//			} else {
//				index = i;
//			}
//			extDuals[cons + index] = tmpSolParts[i];
//		}
//	}
}

void QpExtSolCBC::getExtendedDualFarkas(std::vector<data::QpNum>& extDuals) {
}

void QpExtSolCBC::getExtendedDualFarkas(std::vector<data::QpNum>& extFarkas,
		const data::QpSparseMatrix& constraints,
		const data::QpSparseMatrix& cuts) {
//	unsigned int vars = this->getVariableCount();
//	unsigned int cons = this->getRowCount();
//	std::vector<data::QpNum> boundMultipliers(vars), farkasCertificate;
//	this->getDualFarkas(farkasCertificate);
//	extFarkas = farkasCertificate;
//	extFarkas.resize(cons + 2 * vars, 0);
//	for (unsigned int i = 0; i < boundMultipliers.size(); i++) {
//		boundMultipliers[i].setZero();
//		for (unsigned j = 0; j < constraints[i].size(); j++) {
//			boundMultipliers[i] += (constraints[i][j].value
//					* farkasCertificate[constraints[i][j].index]);
//		}
//		for (unsigned j = 0; j < cuts[i].size(); j++) {
//			boundMultipliers[i] += (cuts[i][j].value
//					* farkasCertificate[cuts[i][j].index]);
//		}
//	}
//	for (unsigned int i = 0; i < boundMultipliers.size(); i++) {
//		if (boundMultipliers[i].isZero())
//			continue;
//		unsigned int index = cons + i;
//		if (boundMultipliers[i] > 0)
//			index += vars;
//		extFarkas[index] = (boundMultipliers[i] *= -1.0);
//	}
}

void QpExtSolCBC::addCut(const std::vector<data::IndexedElement>& lhs,
		data::QpRhs::RatioSign sign, const data::QpNum& rhs) {
//	int size = lhs.size();
//	double lb = -solver->getInfinity();
//	double ub = solver->getInfinity();
//	if (sign == data::QpRhs::equal) {
//		lb = ub = rhs.asDouble();
//	} else if (sign == data::QpRhs::smallerThanOrEqual) {
//		ub = rhs.asDouble();
//	} else {
//		lb = rhs.asDouble();
//	}
//	std::vector<int> indices(size);
//	std::vector<double> values(size);
//	for (unsigned int i = 0; i < size; i++) {
//		indices[i] = lhs[i].index;
//		values[i] = lhs[i].value.asDouble();
//	}
//	solver->addRow(size, indices.data(), values.data(), lb, ub);
}

void QpExtSolCBC::removeCuts() {
//	int num = this->getRowCount() - this->origConstraints;
//	std::vector<int> rows(num);
//	for (unsigned int i = 0; i < num; i++)
//		rows[i] = this->origConstraints + i;
//	solver->deleteRows(num, rows.data());
}

void QpExtSolCBC::removeCut(unsigned int index) {
//	std::vector<int> r(1, index);
//	solver->deleteRows(1, r.data());
}

void QpExtSolCBC::removeCutsFromCut(unsigned int startIndex) {

	if (this->getRowCount() <= startIndex)
		throw utils::ExternSolverException("QpExtSolCLP::removeCut(unsigned int index) --> Index Exception.");

	unsigned int i = 0;
	unsigned int lastIndex = this->getRowCount() - 1;
	unsigned int toDelete = 1+lastIndex-startIndex;
	std::vector<int> iv(toDelete,0);
	while(startIndex!=lastIndex){
		iv[i++]=startIndex++;
	}
	//TODO
}

}
#endif
